/*
 * Upgrade15To16.java
 *
 * Version: $Revision: 1523 $
 *
 * Date: $Date: 2006-05-26 10:23:17 -0400 (Fri, 26 May 2006) $
 *
 * Copyright (c) 2002-2005, Hewlett-Packard Company and Massachusetts
 * Institute of Technology.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * - Neither the name of the Hewlett-Packard Company nor the name of the
 * Massachusetts Institute of Technology nor the names of their
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
package org.dspace.administer;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.sql.SQLException;
import java.io.OutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import org.apache.log4j.Logger;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.cli.MissingOptionException;

import org.dspace.core.PluginManager;
import org.dspace.core.ConfigurationManager;
import org.dspace.content.Bitstream;
import org.dspace.content.BitstreamFormat;
import org.dspace.content.format.DSpaceFormatRegistry;
import org.dspace.content.format.FormatRegistry;
import org.dspace.content.format.FormatHit;
import org.dspace.content.format.FormatIdentifier;
import org.dspace.content.format.FormatConfidence;
import org.dspace.content.format.FormatIdentifierManager;
import org.dspace.core.Context;
import org.dspace.storage.rdbms.DatabaseManager;
import org.dspace.storage.rdbms.TableRow;
import org.dspace.storage.rdbms.TableRowIterator;
import org.dspace.authorize.AuthorizeException;
import org.dspace.content.format.FormatRegistryException;
import org.dspace.content.format.FormatRegistryNotFoundException;
import org.dspace.content.format.FormatRegistryManager;
import org.dspace.content.format.ProvisionalFormatRegistry;
import org.dspace.content.format.DSpaceFormatRegistry;

/**
 * A command-line tool that updates a DSpace 1.5 system to the
 * new BitstreamFormat schema in DSpace 1.6.  It was intended to be
 * run in 4 phases, to allow the DSpace administrator to review the results
 * and ensure they are acceptable before moving on to the next stage.
 *
 * Phase 0: database schema changes.
 * Phase 1: initial format conversion to DSpace (if used) and Provisional
 *           registries.  May be run multiple times.
 * Phase 2: final format conversion, using automatic format identification.
 * Phase 3: finish database schema changes, get rid of now-unused data.
 *
 * Each phase has extensive "sanity checks" to prevent running them out
 * of order or damaging the integrity of the archive.
 *
 * @author Larry Stone
 * @version $Revision: 1523 $
 */
public class Upgrade15To16
{
    /** log4j category */
    private static Logger log = Logger.getLogger(Upgrade15To16.class);

    // -d command option
    private static boolean debug = false;

    // name of this class without package prefix; compute it just in case
    // the classname gets changed, this will have to be updated too.
    private static String myName = null;

    // number of Bitstreams after which to print progress message
    private static final int PROCESS_COUNT = 25000;

    // number of milliseconds after which to print progress message: 15 min.
    private static final int PROCESS_PERIOD = 900 * 1000;

    // put this in "format source" for converted bitstreams.
    private static String phaseOneSource = null;
    private static String phaseTwoSource = null;

    static
    {
        myName = Upgrade15To16.class.getName();
        int lastDot = myName.lastIndexOf(".");
        if (lastDot > 0)
            myName = myName.substring(lastDot+1);

        phaseOneSource = myName + " Phase 1 from ";
        phaseTwoSource = myName + " Phase 2 from ";
    }

    // Set of all format names in stock "DSpace" registry.
    // Need to bypass the plugin manager in case "DSpace" is not configured:
    private static Set<String> dspaceRegistryNames = null;
    static
    {
        try
        {
            dspaceRegistryNames = new DSpaceFormatRegistry().getAllIdentifiers();
        }
        catch (FormatRegistryException e)
        {
            die("Failed getting DSpaceFormatRegistry.getAllIdentifiers(): "+e, null);
        }
    }

    // do we have dspace registry?
    private static boolean hasDSpaceReg = FormatRegistryManager.find(DSpaceFormatRegistry.NAMESPACE) != null;

    // do we have provisional registry?
    private static boolean hasProvisionalReg = FormatRegistryManager.find(ProvisionalFormatRegistry.NAMESPACE) != null;

    // flag to bypass pickier sanity checks
    private static boolean insane = false;

    // exit after showing diagnostic message and/or usage
    private static void die(String msg, Options options)
    {
        if (msg != null)
            System.err.println("ERROR: "+msg+"\n");
        if (options != null)
        {
            HelpFormatter myhelp = new HelpFormatter();
            System.out.println("Convert DSpace 1.5 data model to DSpace 1.6 (renovated BitstreamFormats).");
            myhelp.printHelp("\njava "+Upgrade15To16.class.getName()+" [ -1 | -2 | -3 ] \n", options);
        }
        System.exit(1);
    }

    // true == success
    // check if ready for phase 1 - config should include registries and
    // identifiers, DB has been converted so there is BitstreamFormat
    // table, and Bitstream has a column old_bitstream_format_id that
    // matches the BitstreamFormatRegistry column bitstream_format_id
    private static boolean sanityCheckOne(Context context)
    {
        if (!sanityCheckConfig(context))
            return false;

        try
        {
            // Does Bitstream table have the old_bitstream_format_id column?
            TableRow oc = DatabaseManager.querySingle(context,
                "SELECT COUNT(old_bitstream_format_id) AS count FROM Bitstream;");
            if (debug)
                System.err.println("DEBUG: Got results for COUNT(old_bitstream_format_id): Table="+oc.toString());
            long count = oc.getLongColumn("count");
            if (count <= 0)
            {
                System.err.println("Sanity Check: Unexpected count of column \"old_bitstream_format_id\" in Bitstream; maybe Phase 0 (SQL update) hasn't run yet?");
                return false;
            }
        }
        catch (SQLException e)
        {
            System.err.println("Sanity Check: Failed accessing column \"old_bitstream_format_id\" in Bitstream; maybe Phase 0 (SQL update) hasn't run yet?\n"+e);
            return false;
        }

        try
        {
            // Is there a BitstreamFormat table?  Consider this a
            // success if it returns an answer without getting SQLException:
            TableRow ignore = DatabaseManager.querySingle(context,
                "SELECT COUNT(bitstream_format_id) AS count FROM BitstreamFormat;");
            if (debug)
                System.err.println("DEBUG: Check for new BitstreamFormat table succeeded.");
        }
        catch (SQLException e)
        {
            System.err.println("Sanity Check: Failed accessing BitstreamFormat table, "+
              "maybe Phase 0 (SQL update) hasn't run yet?\n"+e);
            return false;
        }

        return true;
    }

    // check that there has been some attempt to configure Format Registry
    // plugins and Format Identifier plugins.
    private static boolean sanityCheckConfig(Context context)
    {
        boolean result = true;

        // are there FormatRegistry plugins configured?
        String fr[] = PluginManager.getAllPluginNames(FormatRegistry.class);
        if (fr.length < 1)
        {
            System.err.println("Sanity Check: ERROR: No FormatRegistry plugins configured.");
            result = false;
        }

        // are there FormatIdentifier plugins configured?
        Object fi[] = PluginManager.getPluginSequence(FormatIdentifier.class);
        if (fi.length < 1)
        {
            System.err.println("Sanity Check: ERROR: No FormatIdentifier plugins configured.");
            result = false;
        }
        return result;
    }

    // for Report, be sure teh DB is in a state AFTER the SQL conversion
    //  (i.e. there is a BitstreamFormat table), and BEFORE phase 3
    //  (there is a BitstreamFormatRegistry table, and
    //  Bitstream has a column old_bitstream_format_id that matches
    //  the BitstreamFormatRegistry column bitstream_format_id)
    // Returns true if sanity seems intact.
    private static boolean sanityCheckReport(Context context)
    {
        boolean result = sanityCheckConfig(context);

        try
        {
            // Does Bitstream table have the old_bitstream_format_id column?
            TableRow oc = DatabaseManager.querySingle(context,
                "SELECT COUNT(old_bitstream_format_id) AS count FROM Bitstream;");
            if (debug)
                System.err.println("DEBUG: Got results for COUNT(old_bitstream_format_id): Table="+oc.toString());
            long count = oc.getLongColumn("count");
            if (count <= 0)
            {
                System.err.println("Sanity Check: Unexpected count of column \"old_bitstream_format_id\" in Bitstream; maybe Phase 0 (SQL update) hasn't run yet?");
                result = false;
            }
        }
        catch (SQLException e)
        {
            System.err.println("Sanity Check: Failed accessing column \"old_bitstream_format_id\" in Bitstream; maybe Phase 0 (SQL update) hasn't run yet, or Phase 3 has been run already?\n"+e);
            result = false;
        }

        try
        {
            // Is there a BitstreamFormatRegistry table?  Consider this a
            // success if it returns an answer without getting SQLException:
            TableRow ignore = DatabaseManager.querySingle(context,
                "SELECT COUNT(bitstream_format_id) AS count FROM BitstreamFormatRegistry;");
            if (debug)
                System.err.println("DEBUG: Check for new BitstreamFormatRegistry table succeeded.");
        }
        catch (SQLException e)
        {
            System.err.println("Sanity Check: Failed accessing BitstreamFormatRegistry table; maybe SQL update hasn't run yet, or Phase 3 has been run already?\n"+e);
            result = false;
        }
        return result;
    }

    /**
     * For invoking via the command line
     *
     * @param argv
     *            command-line arguments
     */
    public static void main(String[] argv)
    {
        Context context = null;

        Options options = new Options();
        options.addOption("h", "help", false, "Show this help message.");
        options.addOption("n", "dry-run", false, "Run in test mode, do NOT make any changes.");
        options.addOption("d", "debug", false, "enable debugging output");
        options.addOption("I", "insane", false, "Proceed even when some of the pickier sanity checks fail.");
        options.addOption("v", "verbose", false, "Include verbose output, e.g. each Bitstream in --report.");

        // Select the "report" primary operation
        OptionGroup og = new OptionGroup();
        og.addOption(new Option("0", "phase-0", false, "Run phase 0 in the conversion."));
        og.addOption(new Option("1", "phase-1", false, "Run phase 1 in the conversion."));
        og.addOption(new Option("2", "phase-2", false, "Run phase 2 in the conversion. Requires phase 1"));
        og.addOption(new Option("3", "phase-3", false, "Run phase 3 in the conversion. Requires phase 1 and 2."));
        og.addOption(new Option("r", "report", false, "Show report of changes."));
        og.addOption(new Option("p", "provisional", true, "Create file with entries that COULD be added to Provisional registry"));
        og.setRequired(true);
        options.addOptionGroup(og);

        try
        {
            CommandLineParser parser = new PosixParser();
            CommandLine line = parser.parse(options, argv);
             
            if (line.hasOption('h'))
            {
                die(null, options);
            }
             
            boolean dryrun = line.hasOption('n');

            debug = line.hasOption('d');
             
            context = new Context();

            if (line.hasOption("insane"))
                insane = true;

            // needed to modify object model
            context.setIgnoreAuthorization(true);

            // write new Provisional registry, using a temp file
            // in case user chooses to overwrite old registry -- note
            // that this is unwise, since it would destroy any existing
            // entries, but perhaps there aren't any.
            if (line.hasOption('p'))
            {
                File pf = new File(line.getOptionValue('p'));
                File tmpPf = new File(line.getOptionValue('p')+".new");
                PrintStream pout = new PrintStream(new FileOutputStream(tmpPf));
                phaseProvisional(context, pout);
                pout.close();
                tmpPf.renameTo(pf);
            }
            else if (line.hasOption('r'))
            {
                phaseReport(context, line.hasOption('v'));
            }
            else if (line.hasOption('0'))
            {
                phaseZero(context, dryrun, line.hasOption('v'));
            }
            else if (line.hasOption('1'))
            {
                phaseOne(context, dryrun, line.hasOption('v'));
            }
            else if (line.hasOption('2'))
            {
                phaseTwo(context, dryrun, line.hasOption('v'));
            }
            else if (line.hasOption('3'))
            {
                phaseThree(context, dryrun);
            }
            else
                die("Command-line is missing the option to choose which phase, or report.", options);

            context.complete();
            System.exit(0);
        }
        catch (MissingOptionException e)
        {
            die("Missing required command-line option.", options);
        }
        catch (Exception e)
        {
            System.err.println("Exception occurred:" + e);
            e.printStackTrace();

            if (context != null)
            {
                context.abort();
            }

            System.exit(1);
        }
    }

    // KLUDGE to compensate for some old archives which may have
    // "PostScript" as a format name instead of the newer "Postscript"
    // Also some have WordPerfect
    private static String normalizeFormatName(String ns, String id)
    {
        if (ns.equals(DSpaceFormatRegistry.NAMESPACE))
        {
            if (id.equals("PostScript"))
                id = "Postscript";
            else if (id.startsWith("WordPerfect"))
                id = "WordPerfect";
        }
        return id;
    }

    // call BitstreamFormat.findByIdentifier but return null when
    // the identifier is not found..
    private static BitstreamFormat fbiNoFault(Context context, boolean dryrun,
                                              String ns, String id)
        throws SQLException, AuthorizeException, FormatRegistryException
    {
        id = normalizeFormatName(ns, id);

        try
        {
            // try not to change any state on a dry run, i.e. avoid importing formats
            // getLastModified will throw exception if "id" does not exist..
            if (dryrun)
            {
                FormatRegistry reg = FormatRegistryManager.find(ns);
                reg.getLastModified(id);
                return BitstreamFormat.findUnknown(context);
            }
            else
                return  BitstreamFormat.findByIdentifier(context, ns, id);
        }
        catch (FormatRegistryNotFoundException e)
        {
            return null;
        }
    }

    // Run Phase 0 -- apply database changes.
    private static void phaseZero(Context context, boolean dryrun, boolean verbose)
        throws SQLException, IOException
    {

        // sanity check -- have we already run the RDBMS update?
        try
        {
            // Is there a BitstreamFormat table?  Consider this a
            // FAILURE if it returns an answer without getting SQLException:
            TableRow ignore = DatabaseManager.querySingle(context,
                "SELECT COUNT(bitstream_format_id) AS count FROM BitstreamFormat;");
            System.err.println("Sanity Check: BitstreamFormat table already exists, phase Zero must have been run.");
            die("Failed sanity check, stopping.", null);
        }
        catch (SQLException e)
        {
        }

        if (dryrun)
            System.err.println("Would be loading SQL from \"database_schema_15-16.sql\"");
        else
            DatabaseManager.loadSql(getScript("database_schema_15-16.sql"));

        System.err.println("Database has been updated to DSpace 1.6 schema.");
    }

    // show progress statement on long translation, count of BS and
    // elapsed time in mSec
    private static void printProgress(int count, long elapsed)
    {
        elapsed = elapsed / 1000;
        long min = elapsed / 60;
        long hour = min / 60;
        System.err.println(String.format("  Processed %,13d Bitstreams, %d:%02d:%02d hours elapsed.",
                                         count, hour, min %60, elapsed % 60));
    }

    // Run Phase 1 -- convert formats to analogues in DSpace and Provisional reg
    private static void phaseOne(Context context, boolean dryrun, boolean verbose)
        throws SQLException, AuthorizeException, FormatRegistryException

    {
        int total = 0, wasUnknown = 0, nowUnknown = 0, translatedDS = 0,
            translatedP = 0,
            userFormatDescription = 0;
        long startTime = System.currentTimeMillis();

        if (!sanityCheckOne(context))
            die("Failed sanity check, stopping.", null);

        // map of support levels to set later
        Map<BitstreamFormat,Integer> supportLevel = new HashMap<BitstreamFormat,Integer>();

        // Count of all upgradeable BS
        // bitstream_format_id == NULL filters out the unconverted ones.
        TableRow crow = DatabaseManager.querySingle(context,
    "SELECT COUNT(bs.bitstream_id) "+
    "FROM BitstreamFormatRegistry as bsf, Bitstream as bs "+
    "WHERE bs.old_bitstream_format_id = bsf.bitstream_format_id AND bs.deleted = 'f';");
        long allTotal = crow.getLongColumn("count");

        // Check all bitstreams for updated BSF mapping
        // bitstream_format_id == NULL filters out the unconverted ones.
        TableRowIterator tri = DatabaseManager.query(context,
    "SELECT bs.bitstream_id, bsf.short_description, bsf.support_level "+
    "FROM BitstreamFormatRegistry as bsf, Bitstream as bs "+
    "WHERE bs.old_bitstream_format_id = bsf.bitstream_format_id "+
    "  AND bs.bitstream_format_id IS NULL AND bs.deleted = 'f';");
        int lastCount = 0;
        long lastTime = System.currentTimeMillis();
        while (tri.hasNext())
        {
            TableRow row = tri.next();
            String oldName = row.getStringColumn("short_description");
            BitstreamFormat nf;

            ++total;
            if ((total - lastCount) >= PROCESS_COUNT ||
                (System.currentTimeMillis() - lastTime) >= PROCESS_PERIOD)
            {
                lastTime = System.currentTimeMillis();
                lastCount = total;
                printProgress(total, lastTime - startTime);
            }
            int bsid = row.getIntColumn("bitstream_id");
            Bitstream bs = Bitstream.find(context, bsid);
            String ufd = bs.getUserFormatDescription();

            if (oldName.equals("Unknown"))
                ++wasUnknown;

            // 1. If Unknown *AND* UserFormatDescription is set, keep
            // the Unknown format, and set confidence to MANUAL since
            // the format description was manually forced.
            // NOTE: Must restore the UFD, setting Format clears it.
            if (oldName.equals("Unknown") && ufd != null)
            {
                if (dryrun)
                {
                    if (verbose)
                        System.err.println("  BITSTREAM ID="+String.valueOf(bsid)+", name="+bs.getName()+", oldFmt="+oldName+
                            "  ..has UserFormatDescription, would preserve it.");
                }
                else
                {
                    bs.setFormat(BitstreamFormat.findUnknown(context));
                    bs.setUserFormatDescription(ufd);
                    bs.setFormatConfidence(FormatConfidence.MANUAL);
                    bs.setFormatSource(phaseOneSource+"UserFormatDescription");
                }
                ++userFormatDescription;
            }

            // 2. If "DSpace" namespace is available, try to
            //    locate the exact matching format:
            else if (hasDSpaceReg &&
                     (nf = fbiNoFault(context, dryrun,
                       DSpaceFormatRegistry.NAMESPACE, oldName)) != null)
            {
                if (dryrun)
                {
                    if (verbose)
                        System.err.println("  BITSTREAM ID="+String.valueOf(bsid)+", name="+bs.getName()+", oldFmt="+oldName+
                            "  ..has format in DSpace registry, would convert it.");
                }
                else
                {
                    bs.setFormat(nf);
                    bs.setFormatConfidence(FormatConfidence.CIRCUMSTANTIAL);
                    bs.setFormatSource(phaseOneSource+oldName);
                    if (!supportLevel.containsKey(nf))
                        supportLevel.put(nf, new Integer(row.getIntColumn("support_level")));
                }
                ++translatedDS;
            }

            // 3. If "Provisional" namespace is available, try to
            //    locate the exact matching format:
            else if (hasProvisionalReg &&
                     (nf = fbiNoFault(context, dryrun,
                       ProvisionalFormatRegistry.NAMESPACE, oldName)) != null)
            {
                if (dryrun)
                {
                    if (verbose)
                        System.err.println("  BITSTREAM ID="+String.valueOf(bsid)+", name="+bs.getName()+", oldFmt="+oldName+
                            "  ..has format in Provisional registry, would convert it.");
                }
                else
                {
                    bs.setFormat(nf);
                    bs.setFormatConfidence(FormatConfidence.CIRCUMSTANTIAL);
                    bs.setFormatSource(phaseOneSource+oldName);
                    if (!supportLevel.containsKey(nf))
                        supportLevel.put(nf, new Integer(row.getIntColumn("support_level")));
                }
                ++translatedP;
            }
            else
            {
                if (dryrun && verbose)
                    System.err.println("  BITSTREAM ID="+String.valueOf(bsid)+", name="+bs.getName()+", oldFmt="+oldName+
                            "  ..CANNOT BE CONVERTED in Phase 1 (will be guessed)");
            }

            if (!dryrun)
            {
                bs.update();
            }
                 
            // to help the memory footprint..
            context.removeCached(bs, bs.getID());
            
        }
        tri.close();

        // set support level of new BSFs  to values taken from old BSF table
        for (Map.Entry<BitstreamFormat,Integer> e : supportLevel.entrySet())
        {
            BitstreamFormat bsf = e.getKey();
            bsf.setSupportLevel(e.getValue().intValue());
            bsf.update();
        }

        System.err.println("\nPhase One Summary: Bitstream Formats Converted:"+
            "\n  Out of "+String.valueOf(allTotal)+ " upgradable bitstreams in archive,"+
            "\n  Total bitstreams touched = "+String.valueOf(total)+
            "\n  Had old \"Unknown\" format = "+String.valueOf(wasUnknown)+
            "\n  ..of Unknown, touched "+String.valueOf(userFormatDescription)+" with UserFormatDescription set."+
            "\n  Translated to DSpace namespace = "+String.valueOf(translatedDS)+
            "\n  Translated to Provisional namespace = "+String.valueOf(translatedP)+
            "\n");
    }

    // Run Phase 2 -- Automatic format identification of unprocessed bitstreams
    private static void phaseTwo(Context context, boolean dryrun, boolean verbose)
        throws SQLException, IOException, AuthorizeException, FormatRegistryException

    {
        long startTime = System.currentTimeMillis();

        if (!sanityCheckOne(context))
            die("Failed sanity check, stopping.", null);

        // Sanity Check for un-processed Bitstreams with UserFormatDescription
        // If this query finds any rows, Phase One probably wasn't run.
        boolean sanity = true;
        TableRowIterator tri = DatabaseManager.query(context,
  "SELECT bs.bitstream_id, bs.name, bs.user_format_description "+
  "FROM BitstreamFormatRegistry as bsf, Bitstream as bs "+
  "WHERE bs.old_bitstream_format_id = bsf.bitstream_format_id  "+
  "   AND bs.bitstream_format_id IS NULL AND bs.deleted = 'f'"+
  "   AND bsf.short_description = 'Unknown'  "+
  "   AND bs.user_format_description IS NOT NULL ;");
        if (tri.hasNext())
        {
            sanity = false;
            System.err.println("Sanity Check: Found un-converted Bitstreams with UserFormatDescription set, have you run Phase One yet?");
            while (tri.hasNext())
            {
                TableRow row = tri.next();
                System.err.println("  Bitstream ID="+String.valueOf(row.getIntColumn("bitstream_id"))+
                                   ", Name="+row.getStringColumn("name")+
                                   ", UserFormatDescription="+row.getStringColumn("user_format_description"));
            }
        }

        // Sanity Check: if we have DSpace registry configured,
        // look for some Bitstreams with converted BSF IDs.
        // These will be non-Unknown BSs with non-NULL bitstream_format_id
        if (hasDSpaceReg)
        {
            TableRow crow = DatabaseManager.querySingle(context,
  "SELECT COUNT(bs.bitstream_id) "+
  "FROM BitstreamFormatRegistry as bsf, Bitstream as bs "+
  "WHERE bs.old_bitstream_format_id = bsf.bitstream_format_id  "+
  "AND bsf.short_description != 'Unknown' "+
  "AND bs.bitstream_format_id IS NOT NULL AND bs.deleted = 'f';");
            if (debug)
                System.err.println("DEBUG: PhaseTwo Sanity Check: Got results for COUNT(bitstream_id): Table="+crow.toString());
            long count = crow.getLongColumn("count");
            if (count <= 0)
            {
                System.err.println("Sanity Check: No Converted Bitstreams found, even though DSpace registry was configured; have you run Phase One yet?");
                sanity = false;
            }
        }
        if (!sanity)
        {
            if (insane)
                System.err.println("WARNING: Bypassing Phase Two sanity checks and proceeding anyway.");
            else
                die("Failed Phase 2 sanity checks, please try Phase 1 again, or use --insane to force a bypass.", null);
        }

        // Phase 2 processing: Look for Bitstreams that have not
        // been converted yet, and use FormatIdentifier plugins
        // to divine the format.
        int total = 0, wasUnknown = 0, nowUnknown = 0,
            guessed = 0;

        // Try to FormatIdentify all Bitstreams that have either:
        //  a. not yet been processed at all (no new identifier)
        //  b. are still Unknown.
        // Unfortunately there is no mechanism to "improve" the guessing
        // once *some* format is found.  That can be done later with regular 1.6 admin tools.
        tri = DatabaseManager.query(context,
    "SELECT bs.bitstream_id, bsf.short_description, bs.bitstream_format_id "+
    "FROM BitstreamFormatRegistry as bsf, Bitstream as bs "+
    "WHERE bs.old_bitstream_format_id = bsf.bitstream_format_id "+
    "  AND ( bs.bitstream_format_id IS NULL OR "+
    "        ( bs.user_format_description IS NULL AND "+
    "          bs.bitstream_format_id = "+ String.valueOf(BitstreamFormat.findUnknown(context).getID())+" )) "+
    "  AND bs.deleted = 'f';");
        int lastCount = 0;
        long lastTime = System.currentTimeMillis();
        while (tri.hasNext())
        {
            TableRow row = tri.next();
            ++total;
            if ((total - lastCount) >= PROCESS_COUNT ||
                (System.currentTimeMillis() - lastTime) >= PROCESS_PERIOD)
            {
                lastTime = System.currentTimeMillis();
                lastCount = total;
                printProgress(total, lastTime - startTime);
            }
            String oldName = row.getStringColumn("short_description");
            int bsid = row.getIntColumn("bitstream_id");
            Bitstream bs = Bitstream.find(context, bsid);

            // final sanity check, make sure we don't lose UFD.
            if (oldName.equals("Unknown"))
            {
                ++wasUnknown;
                if (bs.getUserFormatDescription() != null)
                    die("Error in Phase 2: Previous phase missed Bitstream #"+String.valueOf(bsid)+
                        ", which has a UserFormatDescription.  Rerun Phase 1.", null);
            }

            // guess the format, and add "phase 2" to the source so we can
            // find these bitstreams for reprocessing later.
            if (dryrun)
            {
                FormatHit h = FormatIdentifierManager.identifyFormat(context, bs);
                if (verbose)
                    System.err.println("  Guessed format of Bitstream ID="+
                      String.valueOf(bsid)+", name="+bs.getName()+", oldfmt="+oldName+
                      ", newFormat="+h.name+" ("+h.getIdentifier()+") conf="+h.confidence+
                      " from "+h.source);
                if (h.confidence != FormatConfidence.UNIDENTIFIED)
                    ++guessed;
                else
                    ++nowUnknown;
            }
            else
            {
                FormatIdentifierManager.identifyAndSetFormat(context, bs);
                bs.setFormatSource(phaseTwoSource+bs.getFormatSource());
                bs.update();
                if (bs.getFormat().isUnknown())
                    ++nowUnknown;
                else
                    ++guessed;
            }

            // to help the memory footprint..
            context.removeCached(bs, bs.getID());
        }
        tri.close();

        System.err.println("\nPhase Two Summary: Bitstream Formats Guessed:"+
            "\n  Total bitstreams touched = "+String.valueOf(total)+
            "\n  Automatically identified = "+String.valueOf(guessed)+
            "\n  Had old \"Unknown\" format = "+String.valueOf(wasUnknown)+
            "\n  Converted to new \"Unknown\" = "+String.valueOf(nowUnknown)+
            "\n");
    }

    // Run Phase 3
    // final phase: drop the tables which are no longer needed.
    // since this is DESTRUCTIVE, be very careful to sanity-check first.
    // ensure ALL bitstreams have a current BitstreamFormat, at least.
    private static void phaseThree(Context context, boolean dryrun)
        throws SQLException, AuthorizeException, FormatRegistryException

    {
        if (!sanityCheckOne(context))
            die("Failed sanity check, stopping.", null);

        // borrow Report's sanity check since it looks for BitstreamFormatRegistry table
        if (!sanityCheckReport(context))
            die("Failed sanity check, stopping.", null);

        // Sanity Check: no un-processed Bitstreams
        // If this query finds any rows, Phase Two probably wasn't run.
        TableRow crow = DatabaseManager.querySingle(context,
            "SELECT COUNT(bitstream_id) FROM Bitstream  "+
            "WHERE bitstream_format_id IS NULL AND deleted = 'f';");
        if (debug)
            System.err.println("DEBUG: PhaseThree Sanity Check: Got results for COUNT(bitstream_id): Table="+crow.toString());
        long count = crow.getLongColumn("count");
        if (count > 0)
        {
            System.err.println("Sanity Check: Some UN-CONVERTED Bitstreams found: have you succeeded in running Phase Two yet?");
            die("Failed sanity check, stopping.", null);
        }

        boolean success = runDDL(context, dryrun, "ALTER TABLE Bitstream DROP COLUMN old_bitstream_format_id;");
        success &= runDDL(context, dryrun, "DROP TABLE FileExtension;");
        success &= runDDL(context, dryrun, "DROP TABLE BitstreamFormatRegistry;");

        // FIXME: these may be PostgreSQL-specific
        success &= runDDL(context, dryrun, "DROP SEQUENCE fileextension_seq;");
        success &= runDDL(context, dryrun, "DROP SEQUENCE bitstreamformatregistry_seq;");

        if (success)
            System.err.println("Phase 3 finished, done with BitstreamFormat conversion.");
        else
            System.err.println("Errors executing Phase 3, check diagnostics and the state of your database.");
    }

    // run a DDL statement and give verbose output if it fails.
    // returns true on success.
    private static boolean runDDL(Context context, boolean dryrun, String ddl)
    {
        if (dryrun)
        {
            System.err.println("Would execute DDL: \""+ddl+"\"");
            return true;
        }
        else
        {
            try
            {
                DatabaseManager.updateQuery(context, ddl);
                return true;
            }
            catch (SQLException e)
            {
                System.err.println("SQL Failed, executing DDL statement: \""+ddl+"\"");
                System.err.println("Exception follows: "+e);
            }
            return false;
        }
    }

    // print report including details of each Bitstream subject to conversion,
    // and a summary of statistics.
    private static void phaseReport(Context context, boolean verbose)
        throws SQLException, AuthorizeException, FormatRegistryException
    {
        if (!sanityCheckReport(context))
            return;

        char colSep = ',';

        int countAll = 0,
            countConverted = 0,
            countUnconverted = 0,
            countUdf = 0;

        String header =
"# BITSTREAM%1$c <BitstreamID>%1$c <Name>%1$c <Old BSF Name>%1$c <New BSF>%1$c <New Confidence>%1$c <New Source>%1$c <User Format Desc.>";
        String reportFmt =
"BITSTREAM%1$c %2$4d%1$c %3$s%1$c %4$s%1$c %5$s%1$c %6$s%1$c \"%7$s\"%1$c %8$s";
        if (verbose)
            System.out.println(String.format(header, colSep));

        // Check all bitstreams for updated BSF mapping
        TableRowIterator tri = DatabaseManager.query(context,
    "SELECT bs.bitstream_id, bs.name, bsf.short_description, bs.old_bitstream_format_id, bs.bitstream_format_id "+
    "FROM BitstreamFormatRegistry as bsf, Bitstream as bs "+
    "WHERE bs.old_bitstream_format_id = bsf.bitstream_format_id AND bs.deleted = 'f';");

        while (tri.hasNext())
        {
            TableRow row = tri.next();
            String oldName = row.getStringColumn("short_description");
            int bsid = row.getIntColumn("bitstream_id");
            Bitstream bs = Bitstream.find(context, bsid);
            int newFormatId = row.getIntColumn("bitstream_format_id");
            String ufd = bs.getUserFormatDescription();

            if (verbose)
                System.out.println(String.format(reportFmt, colSep,
                    bsid,
                    bs.getName(),
                    oldName,
                    (newFormatId >= 0 ? bs.getFormat().getName() : "(NONE)"),
                    bs.getFormatConfidence(),
                    bs.getFormatSource(),
                    (ufd == null ? "" : ufd)));
                 
            // summary counts
            ++countAll;
            if (newFormatId < 0)
                ++countUnconverted;
            else
                ++countConverted;
            if (oldName.equals("Unknown") && ufd != null)
                ++countUdf;

            context.removeCached(bs, bs.getID());
        }
        System.out.println("\nSUMMARY Total upgradeable Bitstreams  = "+countAll);
        System.out.println("SUMMARY Already converted to New BSF  = "+countConverted);
        System.out.println("SUMMARY Not yet converted to New BSF  = "+countUnconverted);
        System.out.println("SUMMARY BS with UserFormatDescription = "+countUdf);
    }

    // print proposed Provisional registry covering all un-converted formats.
    private static void phaseProvisional(Context context, PrintStream out)
        throws SQLException, AuthorizeException, FormatRegistryException
    {
        if (!sanityCheckReport(context))
            return;

        boolean hasProvisionalReg = FormatRegistryManager.find(ProvisionalFormatRegistry.NAMESPACE) != null;
        out.println(
"<?xml version=\"1.0\"?>\n"+
"<dsr:registry xmlns:dsr=\"http://www.dspace.org/xmlns/dspace/formatRegistry\""+
"  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">");
        out.println("\n<!-- WARNING: This document does NOT include any entries which may "+
"\n    already be in your Provisional registry, so please MERGE these"+
"\n    entries with it. -->\n");

        // Look for potential additions to Provisional registry
        int done = 0;
        TableRowIterator tri = DatabaseManager.query(context,
    "SELECT DISTINCT bsf.short_description, bs.old_bitstream_format_id "+
    "FROM BitstreamFormatRegistry as bsf, Bitstream as bs "+
    "WHERE bs.old_bitstream_format_id = bsf.bitstream_format_id "+
    " AND bs.bitstream_format_id is null AND bs.deleted = 'f';");
        while (tri.hasNext())
        {
            TableRow orow = tri.next();
            int bsfid = orow.getIntColumn("old_bitstream_format_id");
            String oldName = orow.getStringColumn("short_description");

            // "normalize" some variants of old names:
            oldName = normalizeFormatName(DSpaceFormatRegistry.NAMESPACE, oldName);

            // skip formats already in the DSpace and Provisional registries.
            // Also skip exceptions like Unknown, License
            if (dspaceRegistryNames.contains(oldName) ||
                oldName.equals("Unknown") ||
                oldName.equals("License") ||
                (hasProvisionalReg &&
                 fbiNoFault(context, true, ProvisionalFormatRegistry.NAMESPACE, oldName) != null))
                continue;
            ++done;

            // dump out the old BSF record as proposed Provisional entry.
            TableRow row = DatabaseManager.find(context, "BitstreamFormatRegistry", bsfid);

            String name = row.getStringColumn("short_description");
            out.println(" <dsr:entry>");
            out.println("  <dsr:name>"+xmlEscape(name)+"</dsr:name>");
            out.println("  <dsr:description>"+xmlEscape(row.getStringColumn("description"))+"</dsr:description>");
            out.println("  <dsr:mime-type>"+row.getStringColumn("mimetype")+"</dsr:mime-type>");
            out.println("  <dsr:external-identifiers>");
            out.println("    <dsr:identifier>"+
                               FormatRegistryManager.makeIdentifier(ProvisionalFormatRegistry.NAMESPACE, name)+
                               "</dsr:identifier>");
            out.println("  </dsr:external-identifiers>");

            // get extensions, if any
            TableRowIterator etri = DatabaseManager.query(context,
                "SELECT extension FROM fileextension "+
                "WHERE bitstream_format_id = ? ;", bsfid);
            if (tri.hasNext())
            {
                out.println("  <dsr:external-signatures>");
                while (etri.hasNext())
                {
                    TableRow erow = etri.next();
                    out.println("    <dsr:extension>"+erow.getStringColumn("extension")+"</dsr:extension>");
                }
                out.println("  </dsr:external-signatures>");
            }
            out.println(" </dsr:entry>");
        }
        if (done == 0)
            out.println("<!-- All Old Format Registry entries are already in DSpace and Provisional registries. -->");
        out.println("</dsr:registry>");
    }

    // returns string with '&', '<', and '>' chars escaped for XML
    private static String xmlEscape(String s)
    {
        StringBuilder result = new StringBuilder(s.length());
        for (int i = 0; i < s.length(); ++i)
        {
            if (s.charAt(i) == '&')
                result.append("&amp;");
            else if (s.charAt(i) == '<')
                result.append("&lt;");
            else if (s.charAt(i) == '>')
                result.append("&gt;");
            else
                result.append(s.charAt(i));
        }
        return result.toString();
    }

    /**
     * Attempt to get the named SQL script, with the following rules:
     * [install]/etc/<db.name>/<name>
     * [install]/etc/<name>
     * [install]/<name>
     * Stolen from InitializeDatabase class.
     */
    private static FileReader getScript(String name)
        throws FileNotFoundException, IOException
    {
        String dbName = ConfigurationManager.getProperty("db.name");
        File myFile = null;
        // File base = new File(ConfigurationManager.getProperty("dspace.dir"));

        if (dbName != null)
        {
            myFile = new File("etc" + File.separator + dbName +
                                     File.separator + name);
            if (myFile.exists())
                return new FileReader(myFile.getCanonicalPath());
        }

        myFile = new File("etc" + File.separator + name);
        if (myFile.exists())
            return new FileReader(myFile.getCanonicalPath());

        return new FileReader(new File(name));
    }
}
