package org.dspace.content.format;

import java.util.Map;
import java.util.HashMap;

/**
 * Values and conversion functions for the "format confidence" metric.
 * This enum indicates the how much the data format identified for a
 * Bitstream is believed to be correct, a factor influenced heavily
 * by the characteristics and mechanism by which it was identified.
 *
 * @author Larry Stone
 * @version $Revision: 1323 $
 */
public enum FormatConfidence
{
    /**
     * No format was identified.  The unknown format is assigned.
     */
    UNIDENTIFIED(0),

    /**
     * A format was found but it was based on "circumstantial evidence",
     * i.e. external properties of the Bitstream such as its name or a
     * MIME type attached to it on ingest.
     */
    CIRCUMSTANTIAL(200),

    /**
     * The data format was determined by coarse examination of the
     * contents of the Bitstream and comparison againt the known
     * characteristics of generic formats such as plain text, or
     * comma-separated-values files.
     */
    HEURISTIC(300),

    /**
     * The format was identified by matching its content positively
     * against an internal signature that describes a "generic"
     * (supertype) format or family of formats.
     */
    POSITIVE_GENERIC(400),

    /**
     * The format was identified by matching its content positively
     * against an internal signature that describes a specific, precise,
     * data format.
     */
    POSITIVE_SPECIFIC(500),

    /**
     * Contents of Bitstream validated as conforming to the identified
     * format, this is the highest confidence of automatic identification.
     */
    VALIDATED(550),

    /**
     * Format was specifically mentioned in metadata supplied at ingest, with
     * the Bitstream.  This implies something more precise than a MIME type.
     */
    INGEST(600),

    /**
     * Format was identified interactively by a user, which acts as an
     * override of automatic format identification.
     */
    MANUAL(700);

    // an enum's integer value
    private final int value;

    // map to translate the number and string versions back to enum values
    private static Map<Integer,FormatConfidence> valueMap = new HashMap<Integer,FormatConfidence>();
    static
    {
        for (FormatConfidence fc : FormatConfidence.values())
        {
            valueMap.put(new Integer(fc.value()), fc);
        }
    }

    /**
     * Required constructor with a numeric argument.
     * @param n the numeric value
     */
    FormatConfidence(int n)
    {
        this.value = n;
    }

    /**
     * Get the numeric value of a FormatConfidence
     * @return the declared integer value of this enum constant.
     */
    public int  value()
    {
        return value;
    }

    /**
     * Get value corresponding to integer, allowing confidence to be
     * stored in e.g. a RDBMS integer column.
     * @param n the numeric value
     * @return the enum corresponding to given integer value, null if out of range.
     */
    public static FormatConfidence valueOf(int n)
    {
        FormatConfidence result = valueMap.get(new Integer(n));
        return (result == null) ? UNIDENTIFIED : result;
    }

    /**
     *  NOTE:  Methods to convert to and from String are automatically
     *  implemented, which are useful for crosswalking to/from e.g. DIM:
     *
     * public String toString()
     * public static FormatConfidence valueOf(String s)
     */
}
