package org.dspace.content.format;

import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

import org.apache.log4j.Logger;

import org.dspace.core.Context;
import org.dspace.core.PluginManager;
import org.dspace.content.Bitstream;
import org.dspace.content.BitstreamFormat;
import org.dspace.authorize.AuthorizeException;

/**
 * Static class that manages the process of identifying formats.
 * <p>
 * @author Larry Stone
 * @version $Revision: 1323 $
 * @see FormatIdentifier
 */
public class FormatIdentifierManager
{
    /** log4j logger */
    private static Logger log = Logger.getLogger(FormatIdentifierManager.class);

    /**
     * Identify all formats matching "target", returning raw hit list.
     * List is guaranteed to contain at least one hit for "unknown" even
     * if all methods fail.
     *
     * @param context the DB context
     * @param target the Bitstream to identify
     * @return  collection of all format identification hits.
     */
    public static List<FormatHit> identifyAllFormats(Context context, Bitstream target)
        throws AuthorizeException, SQLException, IOException
    {
        List<FormatHit> results = new ArrayList<FormatHit>();
        
        for (Object idMethod : PluginManager.getPluginSequence(FormatIdentifier.class))
        {
            try
            {
                results = ((FormatIdentifier)idMethod).identifyFormat(context, target, results);
            }
            catch (FormatIdentifierException fie)
            {
                log.warn(fie);
                boolean isTemp = fie instanceof FormatIdentifierTemporaryException;
                results.add(new FormatHit(FormatRegistryManager.INTERNAL_NAMESPACE,
                                      FormatRegistryManager.UNKNOWN_FORMAT_IDENTIFIER,
                                      FormatRegistryManager.getUnknownFormatName(context),
                                      FormatConfidence.UNIDENTIFIED,
                                      "("+(isTemp ? "TEMPORARY ":"")+
                                      "Failure in "+idMethod.getClass().getName()+")",
                                      (isTemp ? "Temporary failure, please re-try later; ":"")+
                                      "No identification due to exception: "+fie));
            }
        }

        // prune any results that obviously cannot be resolved, i.e. unless
        // the hit is in a configured registry or its identifier is bound:
        for (ListIterator<FormatHit> hi = results.listIterator();
             hi.hasNext(); )
        {
            FormatHit h = hi.next();
            try
            {
                if (!(FormatRegistryManager.find(h.namespace) != null ||
                      BitstreamFormat.findByIdentifier(context, h.namespace, h.identifier, false) != null))
                {
                    if (log.isDebugEnabled())
                        log.debug("Removing un-resolvable hit: "+h);
                    hi.remove();
                }
            }
            catch (FormatRegistryException e)
            {
                log.error("Removing hit that cannot be resolved: ", e);
                hi.remove();
            }
        }


        // add a hit for the Unknown format if nothing else found.
        if (results.isEmpty())
        {
            results.add(new FormatHit(FormatRegistryManager.INTERNAL_NAMESPACE,
                                      FormatRegistryManager.UNKNOWN_FORMAT_IDENTIFIER,
                                      FormatRegistryManager.getUnknownFormatName(context),
                                      FormatConfidence.UNIDENTIFIED,
"(Default after automatic identification failed.)",                                       "Added after automatic identification failed."));
        }

        if (log.isDebugEnabled())
        {
            log.debug("Automatic Format Identification got "+results.size()+" hits.");
            for (FormatHit r : results)
                log.debug("  "+r);
        }
        return results;
    }

    /**
     * Convenience method to identify the format and return only the
     * most relevant "hit".
     *
     * @param context the DB context
     * @param target the Bitstream to identify
     * @return  hit containing single best format identification (never null).
     */
    public static FormatHit identifyFormat(Context context, Bitstream target)
        throws AuthorizeException, SQLException, IOException
    {
        List<FormatHit> hits = identifyAllFormats(context, target);
        return hits.get(0);
    }

    /**
     * Convenience method to encapsulate the whole process of identifying the
     * format of a Bitstream and setting its members to reflect the results.
     * The Bitstream is <i>always</i> modified; even if identification fails,
     * it is set to the Unknown format to reflect that fact.
     *
     * @param context the DB context
     * @param target the Bitstream to identify
     */
    public static void identifyAndSetFormat(Context context, Bitstream target)
        throws AuthorizeException, SQLException, IOException
    {
        try
        {
            FormatHit hit = identifyFormat(context, target);
            hit.applyToBitstream(context, target);
        }
        catch (FormatRegistryException e)
        {
            log.error("Failed while guessing format, recovering.",e);
            target.setFormat(BitstreamFormat.findUnknown(context));
            target.setFormatConfidence(FormatConfidence.UNIDENTIFIED);
            target.setFormatSource("(recovering from "+e+")");
        }
        if (log.isDebugEnabled())
            log.debug("Guessed format for Bitstream ID="+target.getID()+", name="+target.getName()+", format="+target.getFormat().getName());
    }
}
