package org.dspace.content.format;

import org.dspace.content.BitstreamFormat;
import org.dspace.core.Context;
import org.dspace.authorize.AuthorizeException;

import java.net.URL;
import java.util.Date;
import java.sql.SQLException;

/**
 * File/data format registry plugins implement this interface.  This is
 * the primary means of communication between a DSpace archive and the
 * external data format registry.  A DSpace user interface may also have
 * its own way of accessing the registry to obtain identifiers, but since
 * registries vary widely in capabilities and data architecture, it did not
 * make sense to limit UI applications to the lowest-common-denominator
 * interface expressed here.
 * <p>
 * Note that the calls getNamespace() and getContactURL() are intended
 * for UI code to coordinate its registry usage with the core code.
 * <p>
 * The Plugin Manager creates one instance of each registry plugin for the
 * JVM, on demand.  The implementation must manage its own resources, although
 * currently it is never shutdown.
 * <p>
 * This plugin provides a service that is essential to the operation of DSpace.
 * We recommend that registry implementations depend on local (e.g. filesystem)
 * resources as much as possible, for example, local caches of network-
 * accessed resources, to prevent outages from affecting DSpace service any
 * more than necessary -- e.g. if a networked registry is locally cached, then
 * all cached formats would be accessible even though new entries are not.
 *
 * @author Larry Stone
 * @version $Revision: 1323 $
 * @see FormatRegistryManager
 */
public interface FormatRegistry {

    /**
     * NOTE: by convention, the namespace should also be available as
     * public static String class variable named "NAMESPACE" so application
     * code has access to it without instantiating the plugin.
     * Since an interface cannot require static definitions, this must
     * be written as a convention we ask implementors to follow.
     */
    // e.g.: public static String NAMESPACE = "MY_REGISTRY";

    /**
     * Gets the "namespace" of external identifiers managed by this registry.
     * This must be a globally unique name amongst registries, and
     * should consist of alphanumeric characters only.  Use the proper name
     * of the external registry represented by the plugin, e.g.
     * "PRONOM", "GDFR", "LOC", etc.
     *
     * @return  the DSpace namespace of this registry.
     */
    public String getNamespace();

    /**
     * Return an URL intended to be the contact address or other top-level
     * configuration information required by the underlying format
     * registry implementation. It might be a network resource or local file.
     * This method allows other code using the same external registry within
     * the DSpace server (e.g. specialized UI components) to contact the
     * registry the same way that the plugin does.
     *
     * @return  URL for the main contact point of the registry resource;
     *          may be network resource or a local file.
     */
    public URL getContactURL();

    /**
     * Import a new data format from the registry, creates a new
     * BitstreamFormat.  There must not be any existing BSF with the same
     * namespace and identifier.  The implementation is expected to set
     * the new BSF's name, description, MIME type, canonical filename
     * extension, synonym identifiers, and last-import date.
     *
     * It is also expected to set the supportLevel, possibly to a default
     * value; ideally, the initial value is dictated by configurable policy.
     *
     * NOTE: This is called in the user's privilege context, and most users
     * are not authorized to create new BitstreamFormats.  Therefore, the
     * implementation must call FormatRegistryManager.createBitstreamFormat()
     * to create the new BSF.  Also, it MUST NOT call bsf.update(), the
     * manager code does that for it at an elevated privilege level.
     *
     * DSpace code should therefore never call this method directly, but only
     * through the FormatRegistryManager dispatch method.
     *
     * @param context the database context
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return the new BSF, or null if the external format cannot befound.
     * @throws FormatRegistryException for any failure contacting or accessing the registry.
     */
    public BitstreamFormat importExternalFormat(Context context, String identifier)
        throws FormatRegistryException, AuthorizeException, SQLException;

    /**
     * Gets all external identifiers known to be synonyms of the given
     * identifier, in this registry.  The synonyms must be in DSpace namespaced
     * external identifier format, since there may be identifiers in other registries -- e.g.
     * the built-in DSpace registry may include PRONOM or GDFR synonyms.
     * The list may include the identifier in the query.
     * Use FormatRegistryManager.makeIdentifier() to create them.
     *
     * @param context the database context
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return array of all synonyms for this format in the registry.
     */
    public String[] getSynonymIdentifiers(Context context, String identifier)
        throws FormatRegistryException, AuthorizeException;

    /**
     * Re-import the file format metadata from the external registry,
     * updating all values of the corresponding BSF except for those which
     * have been marked as "override".  These include
     * new BSF's name, description, MIME type, and canonical filename
     * extension.  It should also add any new synonym identifiers, and
     * and set the last-import date.  Unlike the import function, it must
     * call the update() method on the BSF, which will fail unless the
     * user in the current Context is an Administrator.
     * Although users can be expected to invoke operations which result in
     * new BSFs being imported, the update operation is only available to
     * administrators.
     *
     * @param context the database context
     * @param bsf the BitstreamFormat to update
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @throws FormatRegistryException for any failure contacting or accessing the registry, or if the identifier cannot be found.
     */
    public void updateBitstreamFormat(Context context, BitstreamFormat existing, String identifier)
        throws FormatRegistryException, AuthorizeException, SQLException;
    
    /**
     * Get the time when the given registry entry was last changed.  May
     * return null when this information is not available.
     *
     * @param identifier the bare (NON-namespaced) external identifier of the format
     * @return date/time when this entry was last changed, or null if not available.
     * @throws FormatRegistryException upon failure accessing the registry.
     */
    public Date getLastModified(String identifier)
        throws FormatRegistryException;

    /**
     * Predicate, true if format named by sub is a subtype or
     * otherwise "conforms" to the format defined by fmt -- this means
     * Bitstreams identified as <i>subtype</i> would also be identified as
     * <i>fmt</i>.
     *
     * @param sub bare (NON-namespaced) external identifier of the subtype format
     * @param fmt bare (NON-namespaced) external identifier of the supertype format
     * @throws FormatRegistryException upon failure accessing the registry.
     */
    public boolean conformsTo(String sub, String fmt)
        throws FormatRegistryException;
    
    /**
     * Free any resources associated with this registry connection,
     * since it will not be used any more.
     */
    public void shutdown();
}
