package org.dspace.content.format;

import org.dspace.content.BitstreamFormat;
import org.dspace.core.Context;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.Email;
import org.dspace.core.PluginManager;
import org.dspace.core.I18nUtil;
import org.dspace.authorize.AuthorizeException;

import java.net.URL;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.Set;
import org.apache.log4j.Logger;

/**
 * Factory class for FormatRegistry plugins, this dispatches all registry
 * operations to the correct implementation.  It also supplies convenience
 * methods to manipulate external format identifiers.  Finally, it contains
 * the implementation of the "Internal" registry that is required by every
 * DSpace archive.  It is embedded here because it only provides the Unknown
 * format and no other code needs access to it.
 *
 * @author Larry Stone
 * @version $Revision: 1323 $
 * @see FormatRegistry
 */
public class FormatRegistryManager
{
    /** Namespace of Internal format registry. */
    public static final String INTERNAL_NAMESPACE = "Internal";

    /** Identifier of "Unknown" format (in Internal registry), the only required format */
    public static final String UNKNOWN_FORMAT_IDENTIFIER = "Unknown";


    /** Human-readable name, desc of "Unknown" format, the only required format.
     ** Note they are implemented as keys to the i18n message catalog.
     **/
    private static final String UNKNOWN_NAME_KEY = ".unknown";
    private static final String UNKNOWN_DESC_KEY = ".unknownDescription";

    /** MIME type to use when none is available. */
    public static final String DEFAULT_MIME_TYPE = "application/octet-stream";

    /** log4j logger */
    private static Logger log = Logger.getLogger(FormatRegistryManager.class);

    // implement the Internal format registry, home of Unknown format only.
    public static class InternalFormatRegistry
        implements FormatRegistry
    {
        private final String aliases[] = { INTERNAL_NAMESPACE };

        public String [] getPluginNames()
        {
            return aliases;
        }

        public String getNamespace()
        {
            return INTERNAL_NAMESPACE;
        }

        public URL getContactURL()
        {
            return null;
        }

        public String[] getSynonymIdentifiers(Context context, String identifier)
            throws FormatRegistryException, AuthorizeException
        {
            return new String[0];
        }

        /**
         * Import a new data format - returns a BitstreamFormat.  There
         * must not be any existing BSF with the same namespace and identifier.
         *
         * @return new format or null if identifier not found in registry
         */
        public BitstreamFormat importExternalFormat(Context context, String id)
            throws FormatRegistryException, AuthorizeException, SQLException
        {
            if (!id.equalsIgnoreCase(UNKNOWN_FORMAT_IDENTIFIER))
                throw new FormatRegistryNotFoundException(INTERNAL_NAMESPACE+" format namespace only contains Unknown format, asked for id="+id);

            BitstreamFormat bsf = createBitstreamFormat(context);
            bsf.setName(FormatRegistryManager.getUnknownFormatName(context));
            bsf.setDescription(
                FormatRegistryManager.getMessageCatalog(context).getString(FormatRegistryManager.class.getName()+UNKNOWN_DESC_KEY));
            bsf.setMIMEType(DEFAULT_MIME_TYPE);
            // be sure the canonical ext. ID is first
            bsf.addIdentifier(FormatRegistryManager.makeIdentifier(getNamespace(), UNKNOWN_FORMAT_IDENTIFIER));
            bsf.setSupportLevel(BitstreamFormat.UNKNOWN);
            return bsf;
        }

        public void updateBitstreamFormat(Context context, BitstreamFormat bsf, String id)
            throws FormatRegistryException, AuthorizeException, SQLException
        {
            if (!id.equalsIgnoreCase(UNKNOWN_FORMAT_IDENTIFIER))
                throw new FormatRegistryNotFoundException(INTERNAL_NAMESPACE+" format namespace only contains Unknown format, asked for id="+id);

            if (!bsf.isOverrideName())
                bsf.setName(FormatRegistryManager.getUnknownFormatName(context));
            if (!bsf.isOverrideDescription())
                bsf.setDescription(
                  FormatRegistryManager.getMessageCatalog(context).getString(FormatRegistryManager.class.getName()+UNKNOWN_DESC_KEY));
            if (!bsf.isOverrideMIMEType())
                bsf.setMIMEType(DEFAULT_MIME_TYPE);
            bsf.setLastImported(getNamespace(), UNKNOWN_FORMAT_IDENTIFIER, new Date());
        }

        public Date getLastModified(String identifier)
            throws FormatRegistryException
        {
            return null;
        }

        public boolean conformsTo(String sub, String fmt)
            throws FormatRegistryException
        {
            return sub.equalsIgnoreCase(fmt);
        }

        public void shutdown()
        {
        }
    }

    // instance for the Internal registry "plugin"
    private static FormatRegistry internalReg = new InternalFormatRegistry();

    // cached message catalog - decache if locale changes.
    private static ResourceBundle msgCat = null;

    // cache of last locale, if any
    private static Locale msgLocale = null;

    // get message catalog corresponding to current Locale
    private static ResourceBundle getMessageCatalog(Context context)
    {
        Locale locale = context.getCurrentLocale();
        if (msgCat == null ||
            msgLocale == null || !msgLocale.equals(locale))
        {
            msgLocale = locale;
            msgCat = ResourceBundle.getBundle("Messages", msgLocale);
        }
        return msgCat;
    }

    /**
     * Get Registry implementation corresponding to a Namespace.
     * @return  registry plugin for external format identifier namespace
     */
    public static FormatRegistry find(String namespace)
    {
        // special case for the internal registry, because it must ALWAYS
        // be available.  hardcoding it here protects it from configuration errors.
        if (namespace.equalsIgnoreCase(INTERNAL_NAMESPACE))
            return internalReg;
        else
            return (FormatRegistry)PluginManager.getNamedPlugin(FormatRegistry.class, namespace);
    }

    /**
     * Get ALL format namespaces present in the DSpace data model; this
     * may include artifacts from registries that are no longer configured.
     * @param context the db context
     * @return  array of all available Namespace strings
     */
    public static String[] getAllNamespaces(Context context)
        throws SQLException, AuthorizeException
    {
        String plugins[] = PluginManager.getAllPluginNames(FormatRegistry.class);
        Set<String> all = new HashSet<String>();
        all.add(INTERNAL_NAMESPACE);
        for (String pname : plugins)
            all.add(pname);

        // sanity check against list of NS's from the DB:
        for (String dbname : BitstreamFormat.getAllExternalNamespaces(context))
        {
            if (!all.contains(dbname))
            {
                all.add(dbname);
                log.warn("Found BitstreamFormat External Identifier Namespace \""+dbname+"\" in DB, but there is no plugin for it! May be an artifact.");
            }
        }
        return all.toArray(new String[all.size()]);
    }

    /**
     * Returns _only_ the namespaces of active, configured external registries.
     * Can be cached, since plugin list is fixed at configuration time.
     * @return  array of all available Nameixespace strings
     */
    private static String cachedRegistryNamespaces[] = null;
    public static String[] getRegistryNamespaces()
    {
        if (cachedRegistryNamespaces == null)
        {
            String plugins[] = PluginManager.getAllPluginNames(FormatRegistry.class);
            cachedRegistryNamespaces = new String[plugins.length + 1];
            for (int i = 0; i < plugins.length; ++i)
                cachedRegistryNamespaces[i] = plugins[i];
            cachedRegistryNamespaces[plugins.length] = INTERNAL_NAMESPACE;
        }
        return cachedRegistryNamespaces;
    }

    /**
     * Binds the given external format identifier to a local BitstreamFormat,
     * importing the external format record to create a new one if necessary.
     * If there is already a BSF bound to the identifier or one of its known
     * synonyms, the external identifier (and its synonyms) are added as
     * more synonyms, and that BSF is returned.
     *
     * @param context  the db context
     * @param namespace namespace of external identifier
     * @param identifier identifier of external identifier
     * @return  a new BitstreamFormat, or null if format cannot be found.
     * @throws FormatRegistryException when there is an error accessing the registry.
     */
    public static BitstreamFormat importExternalFormat(Context context, String namespace, String identifier)
        throws FormatRegistryException, AuthorizeException, SQLException
    {
        FormatRegistry r = find(namespace);
        if (r == null)
            throw new FormatRegistryNotFoundException("No registry configured for BitstreamFormat namespace="+namespace);

        // check for any synonyms -- also try looking up given identifier as sanity check.
        BitstreamFormat result = BitstreamFormat.findByIdentifier(context, namespace, identifier, false);
        String synIDs[] = r.getSynonymIdentifiers(context, identifier);
        for (String nsid : synIDs)
        {
            BitstreamFormat synBsf = BitstreamFormat.findByIdentifier(context,
                FormatRegistryManager.namespaceOf(nsid),
                FormatRegistryManager.identifierOf(nsid), false);
            if (synBsf != null)
            {
                if (result == null)
                    result = synBsf;
         
                // if synonyms match two *different* existing BSFs, we have trouble!
                else if (!result.equals(synBsf))
                    throw new FormatRegistryException(
                      "Cannot import the format entry "+
                      makeIdentifier(namespace, identifier)+
                      ", because it has synonym identifiers matching two different, existing "+
                      "BitstreamFormats: "+result.toString()+" and "+synBsf.toString()+
                      ". This condition is not allowed.");
            }
        }
         
        // if we found EXACTLY ONE synonym BSF, add our IDs to it and return it!
        if (result != null)
        {
            result.addIdentifier(namespace,identifier);
         
            // add all synonyms that are not already bound
            for (String nsid : synIDs)
                result.addIdentifier(nsid);

            result.update();
            return result;
        }
         
        // else go ahead and import a new format from registry, and notify
        result = r.importExternalFormat(context, identifier);
        if (result != null)
            bsfUpdate(context, result);

        // send mail, possibly with warnings.
        sendFormatAlert(context, result, makeIdentifier(namespace, identifier));

        return result;
    }

    /**
     * Updates a BSF's cached metadata from the external registry's
     * format metadata, except where properties are set to "override".
     * If a valid "last modified" date is available, skip updating entries
     * when the external registry has not changed since the last update, unless
     * the force parameter is true.
     *
     * @param context  the db context
     * @param bsf the BitstreamFormat to be updated
     * @param namespace namespace of external identifier
     * @param identifier identifier of external identifier
     * @param force flag, if true update anyway whether or not source changed.
     * @throws FormatRegistryException when there is an error accessing the registry.
     */
    public static void updateBitstreamFormat(Context context, BitstreamFormat bsf, String namespace, String identifier, boolean force)
        throws FormatRegistryException, AuthorizeException, SQLException
    {
        FormatRegistry r = find(namespace);
        if (r == null)
            throw new FormatRegistryException("No registry found for namespace="+namespace);

        // check whether update is necessary
        // only if the registry has been modified since BSF was last updated:
        if (!force)
        {
            Date blm = bsf.getLastImported(namespace, identifier);
            Date rlm = r.getLastModified(identifier);
            if (blm != null && rlm != null && rlm.before(blm))
                return;
        }
        r.updateBitstreamFormat(context, bsf, identifier);
        bsf.update();
    }

    /**
     * Calls apropriate registry plugin to compare two namespaced
     * identifiers (which must be in the same namespace).
     * Algorithm: lookup the target format, find all other synonym identifiers,
     * since the registry-dependent conformsTo can only be done within a
     * registry.
     *
     * FIXME:  Perhaps there ought to be a way to indicate when the
     * answer is really "Cannot know" e.g. when there are no IDs from
     * the same registry to compare.  Currently it returns false.
     *
     * @param context database context
     * @param sub namespaced identifier of format that would be subclass.
     * @param fmt namespaced identifier of format which should be superclass.
     * @return true if sub "conforms" to fmt, i.e. sub is the same or subclass
     */
    public static boolean conformsTo(String sub, String fmt)
        throws FormatRegistryException
    {
        String subNs = namespaceOf(sub);
        String subId = identifierOf(sub);
        FormatRegistry r = find(subNs);
        if (r != null)
        {
            String fmtNs = namespaceOf(fmt);
            if (fmtNs.equals(subNs))
                return r.conformsTo(subId, identifierOf(fmt));
            log.warn("ConformsTo unanswerable because namespaces are different: Could not actually compare \""+sub+"\" conforms to \""+fmt+"\"");
        }
        else
            log.warn("ConformsTo unanswerable, no registry for namespace: Could not actually compare \""+sub+"\" conforms to \""+fmt+"\"");
        return false;
    }

    /**
     * Creates a namespaced identifier out of separate namespace and
     * registry-specific identifier.  They are joined into a single
     * string by an opaque process which is undo-able by the
     * namespaceOf() and identifierOf() methods below; this is intentional
     * so that the exact pattern of the merged identifier may be changed
     * later.
     *
     * For the initial implementation, the pattern is simply
     *  {namespace}:{identifier}, i.e. separated with a colon.
     * Fast and simple, but there is some prospect for confusion..
     *
     * FIXME: Consider using a real URI, perhaps in "info:" scheme, e.g.
     *  "info:/dspace/format/<namespace>#<ident>
     */
    public static String makeIdentifier(String namespace, String identifier)
    {
        return namespace + ":" + identifier;
    }

    /**
     * Get the namespace from a merged External Format Identiifer.
     * @return  the namespace portion of a namespaced identifier, or empty str.
     */
    public static String namespaceOf(String nsIdentifier)
    {
        int colon = nsIdentifier.indexOf(':');
        if (colon > 0)
            return nsIdentifier.substring(0, colon);
        return "";
    }

    /**
     * Get the Identifier from a merged External Format Identiifer.
     * @return  the identifier portion of a namespaced identifier or empty str.
     */
    public static String identifierOf(String nsIdentifier)
    {
        int colon = nsIdentifier.indexOf(':');
        if (colon >= 0)
            return nsIdentifier.substring(colon+1);
        else
            return "";
    }

    /**
     * Get the localized, human-readable name of the Unknown format.
     * It is a method so it can be localized.
     * @param context database context
     * @return  the localized name of the Unknown format
     */
    public static String getUnknownFormatName(Context context)
    {
        return getMessageCatalog(context).getString(FormatRegistryManager.class.getName()+UNKNOWN_NAME_KEY);
    }

    /**
     * Send an alert to the designated "alert recipient" - i.e. the
     * application administrator, announcing a new format was created by import.
     * <P>
     * The recipient is configured via the "alert.recipient" property in
     * <code>dspace.cfg</code>. If this property is omitted, no alerts are
     * sent.
     * <P>
     * This method buries exceptions since a failure to send mail must not
     * break the new format creation; notification is not critical.
     *
     * @param context database context
     * @param bsf the BitstraemFormat just imported
     * @param nsid namespaced external identifier of the format imported
     */
    private static void sendFormatAlert(Context context, BitstreamFormat bsf, String nsid)
    {
        String subjWarn = "";
        String warning = "";
        String mime = bsf.getMIMEType();
        int supportLevel = bsf.getSupportLevel();
        ResourceBundle mc = getMessageCatalog(context);
        String prefix = FormatRegistryManager.class.getName();

        try
        {
            // add warning if format is missing distinct MIME type
            if (mime.equals(DEFAULT_MIME_TYPE))
            {
                subjWarn = mc.getString(prefix+".subjectWarning")+" ";
                warning += mc.getString(prefix+".mimeWarning") + "\n\n";
            }
             
            // add warning when support level is UNKNOWN
            if (supportLevel == BitstreamFormat.UNKNOWN)
            {
                subjWarn = mc.getString(prefix+".subjectWarning")+" ";
                warning += mc.getString(prefix+".unsupportedWarning") + "\n\n";
            }

            String recipient = ConfigurationManager.getProperty("alert.recipient");
            if (recipient != null)
            {
                Email email = ConfigurationManager.getEmail(I18nUtil.getEmailFilename(context.getCurrentLocale(), "import_format"));
                email.setSubject(subjWarn+
                                 mc.getString(prefix+".subjectMiddle")+
                                 " "+bsf.getName());
                email.addRecipient(recipient);
                // Parameters:
                /* 0 */ email.addArgument(warning);
                /* 1 */ email.addArgument(String.valueOf(bsf.getID()));
                /* 2 */ email.addArgument(mc.getString(prefix+"."+BitstreamFormat.supportLevelText[supportLevel].toLowerCase()));
                /* 3 */ email.addArgument(nsid);
                /* 4 */ email.addArgument(bsf.getName());
                /* 5 */ email.addArgument(mime);
                /* 6 */ email.addArgument(bsf.getDescription());
                email.send();
            }
        }
        catch (Exception e)
        {
            // Not much we can do here!
            log.error("Unable to send email format alert:", e);
        }
    }

    /**
     * Implementations of FormatRegistry.importExternalFormat() call this
     * to create a new internal BitstreamFormat, because it does not require
     * Administrator access, and new formats are usually created in the
     * course of applying the results of automatic format identification.
     * The identifier returns an external format identifier, which is then
     * imported if necessary to obtain the BSF to apply to the Bitstream.
     * <p>
     * This is nothing more than BitstreamFormat.create wrapped by code
     * to suspend authorization checking.
     *
     * @param context database context
     * @return the created BSF (never null)
     */
    public static BitstreamFormat createBitstreamFormat(Context context)
        throws SQLException, AuthorizeException
    {
        boolean oldIA = context.ignoreAuthorization();
        try
        {
            context.setIgnoreAuthorization(true);
            return BitstreamFormat.create(context);
        }
        finally
        {
            context.setIgnoreAuthorization(oldIA);
        }
    }

    // Just like createBitstreamFormat, call bsf.update() without
    // authorization checking -- but this is only used internally, to
    // send the BSF created by the import process to the RDBMS.
    private static void bsfUpdate(Context context, BitstreamFormat bsf)
        throws SQLException, AuthorizeException
    {
        boolean oldIA = context.ignoreAuthorization();
        try
        {
            context.setIgnoreAuthorization(true);
            bsf.update();
        }
        finally
        {
            context.setIgnoreAuthorization(oldIA);
        }
    }
}
