package org.dspace.content.format;

import org.dspace.content.BitstreamFormat;
import org.dspace.content.Bitstream;
import org.dspace.core.ConfigurationManager;
import org.dspace.core.SelfNamedPlugin;
import org.dspace.core.Context;
import org.dspace.authorize.AuthorizeException;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;

import org.apache.log4j.Logger;

/**
 * Text Heuristic File Format Identifier
 *
 * This code borrows heavily from the module ascmagic.c in Ian Darwin's
 * open-source file(1) command, version 4.21.  Authors of that code are:
 *  Ian F. Darwin
 *  Christos Zoulas
 *  Eric Fischer <enf@pobox.com>
 *  Joerg Wunsch <joerg@freebsd.org>
 *
 * It only identifies the specific text file types below.  The DSpace
 * Configuration (or defaults) for each type dictates how it gets
 * translated into one or more external format identifiers to put in
 * FormatHits.  Since a site can configure its own set of registries, it
 * may also need to map the text file formats into identifiers in those
 * registries.
 *
 *  TEXT_ASCII      - plain old 7-bit ASCII
 *  TEXT_ISO8859    - ISO-8859 Latin-1 text
 *  TEXT_EXTENDED   - one of the other 8-bit text encodings (MS, apple?)
 *  TEXT_UTF8       - Unicode in UTF8
 *  TEXT_UTF16      - Unicode in UTF16
 *
 *  Configuration file entries are e.g.
 * formatIdentifier.TextHeuristicIdentifier.formats.ascii = DSpace:Text, PRONOM:x-fmt/22
 * formatIdentifier.TextHeuristicIdentifier.formats.iso8859 = DSpace:Text, PRONOM:x-fmt/283
 * formatIdentifier.TextHeuristicIdentifier.formats.extended = DSpace:Text, PRONOM:x-fmt/283
 * formatIdentifier.TextHeuristicIdentifier.formats.utf8 = DSpace:Text, PRONOM:x-fmt/16
 * formatIdentifier.TextHeuristicIdentifier.formats.utf16 = DSpace:Text, PRONOM:x-fmt/16
 * formatIdentifier.TextHeuristicIdentifier.formats.tar = PRONOM:x-fmt/265
 *
 * <p>
 * @author Larry Stone
 * @version $Revision: 1323 $
 */
public class TextHeuristicIdentifier
    implements FormatIdentifier
{
    /** log4j logger */
    private static Logger log = Logger.getLogger(TextHeuristicIdentifier.class);

    // list of all active registries for quick comparison checks.
    private static String configuredRegistries[] = FormatRegistryManager.getRegistryNamespaces();
    static { Arrays.sort(configuredRegistries); }

    private class Result
    {
        protected String name;
        protected String text;
        protected String identifiers[];

        protected Result(String name, String text, String defaultIdentifiers)
        {
            super();
            this.name = name;
            this.text = text;
            String rawNsids = ConfigurationManager.getProperty(
                "formatIdentifier.TextHeuristicIdentifier.formats."+name);
            if (rawNsids == null)
                rawNsids = defaultIdentifiers;
            this.identifiers = rawNsids.trim().split("\\s*,\\s*");
        }
    }

    // table of details about result codes
    private static Result result[] = null;

    /** Result codes - WARNING, if you add a result code, update init() below! */
    private static final int TEXT_ASCII    = 0; // plain old 7-bit ASCII
    private static final int TEXT_ISO8859  = 1; // ISO-8859 Latin-1 text
    private static final int TEXT_EXTENDED = 2; // one of the other 8-bit text encodings (MS, apple?)
    private static final int TEXT_UTF8     = 3; // Unicode in UTF8
    private static final int TEXT_UTF16    = 4; // Unicode in UTF16
    private static final int TEXT_TAR      = 5; // Unix Tape Archive (tar)

    // WARNING, this must insert a row in result[] for every result code.
    private void init()
    {
        if (result == null)
        {
            result = new Result[6];
            result[TEXT_ASCII]    = new Result("ascii",    "ASCII Text",          "DSpace:Text, PRONOM:x-fmt/22, PRONOM:x-fmt/111");
            result[TEXT_ISO8859]  = new Result("iso8859",  "ISO8859 Text",        "DSpace:Text, PRONOM:x-fmt/283");
            result[TEXT_EXTENDED] = new Result("extended", "Extended ASCII Text", "DSpace:Text, PRONOM:x-fmt/283");
            result[TEXT_UTF8]     = new Result("utf8",     "UTF8 Text",           "DSpace:Text, PRONOM:x-fmt/16");
            result[TEXT_UTF16]    = new Result("utf16",    "UTF16 Text",          "DSpace:Text, PRONOM:x-fmt/16");
            result[TEXT_TAR]      = new Result("tar",      "UNIX Tape Archive",   "PRONOM:x-fmt/265");
        }
    }

    static private Set<String> allIdentifiers = null;

    /**
     * Returns array of all external identifiers of formats recognized
     * by this identifier.  Used by subclass identifier method to detect
     * text formats from which can deduce a subtype.
     * @returns array including all unique format identifiers this can identify.
     */
    protected Set<String> getTextFormatIdentifiers()
    {
        init();
        if (allIdentifiers == null)
        {
            allIdentifiers = new HashSet<String>();
            for (Result r : result)
            {
                for (String id : r.identifiers)
                    allIdentifiers.add(id);
            }
        }
        return allIdentifiers;
    }

    /** size of input buffer, code borrowed from file(1). */
    private static final int HOWMANY = 256 * 1024;

    /** key values for character (byte) identification array */
    private static final int F = 0;  /* character never appears in text */
    private static final int T = 1;  /* character appears in plain ASCII text */
    private static final int I = 2;  /* character appears in ISO-8859 text */
    private static final int X = 3;  /* character appears in non-ISO extended ASCII (Mac, IBM PC) */

    /** Array of 256 flags, one for each char (byte) value. */
    private static final int text_chars[] =
    {
        /*                  BEL BS HT LF    FF CR    */
        F, F, F, F, F, F, F, T, T, T, T, F, T, T, F, F,  /* 0x0X */
        /*                              ESC          */
        F, F, F, F, F, F, F, F, F, F, F, T, F, F, F, F,  /* 0x1X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x2X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x3X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x4X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x5X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x6X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, F,  /* 0x7X */
        /*            NEL                            */
        X, X, X, X, X, T, X, X, X, X, X, X, X, X, X, X,  /* 0x8X */
        X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X,  /* 0x9X */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xaX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xbX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xcX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xdX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xeX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I   /* 0xfX */
    };

    /**
     * Identify format of "target", add hits to "results".
     * <p>
     * @param context the DB context
     * @param target the Bitstream to identify
     * @param results List of results produced by previous identifyFormat methods
     * @return list of results, possibly modified or replacing the result param
     * @throws FormatIdentifierException when a fatal error in the process
     *   of identifying the format prevents it from returning conclusive results
     */
    public List<FormatHit> identifyFormat(Context context, Bitstream target, List<FormatHit> results)
        throws FormatIdentifierException, AuthorizeException, SQLException, IOException
    {
        init();
        int format = identifyFormatInternal(target);
        if (format < 0)
            return results;
        if (format >= result.length)
            throw new FormatIdentifierException("Got intermediate result out of range, result="+format+", bitstream="+target);

        if (log.isDebugEnabled())
            log.debug("Got heuristic result="+result[format].name+" for Bitstream="+target.getName());

        String myname = this.getClass().getName();
        for (String nsid : result[format].identifiers)
        {
            String ns = FormatRegistryManager.namespaceOf(nsid);
            if (Arrays.binarySearch(configuredRegistries, ns) >= 0)
                new FormatHit(ns,
                              FormatRegistryManager.identifierOf(nsid),
                              result[format].text,
                              FormatConfidence.HEURISTIC,
                              myname,
                              null).addToResults(results);
        }
        return results;
    }

    // try heuristics to identify contents of stream.
    private int identifyFormatInternal(Bitstream target)
         throws IOException, SQLException, AuthorizeException
    {
        byte buf[] = new byte[HOWMANY];
        InputStream ibs = target.retrieve();
        int bufLen = ibs.read(buf);
        ibs.close();
        if (bufLen <= 0)
            return -1;

        // First, build histogram of types of bytes
        int kind[] = new int[4];
        Arrays.fill(kind, 0);
        for (int i = 0; i < bufLen; ++i)
        {
            ++kind[text_chars[((int)buf[i]) & 0xff]];
        }

        // first check for special characteristics of tar file, since
        // those can also look like pure ascii.
        if (looks_like_tar(buf, bufLen))
            return TEXT_TAR;

        // it's straight ascii if all characters are in set T
        else if (kind[T] == bufLen)
            return TEXT_ASCII;

        // utf8
        else if (kind[F] == 0 && looks_like_utf8(buf, bufLen))
            return TEXT_UTF8;

        // utf16
        else if (kind[F] == 0 && looks_like_utf16(buf, bufLen))
            return TEXT_UTF16;

        // iso8859 (latin1), all chars are T and I
        else if ((kind[T] + kind[I]) == bufLen)
            return TEXT_ISO8859;

        // extended, all chars are T , X, and I
        else if ((kind[T] + kind[I] + kind[X]) == bufLen)
            return TEXT_EXTENDED;
        else
            return -1;
    }

    // Note: this ASSUMES that buf does NOT contain all 7-bit chars,
    // AND that the 7-bit chars it DOES contain are NOT weird controls.
    // we tested that already with kind[] histogram.
    private boolean looks_like_utf8(byte buf[], int bufLen)
    {
        for (int i = 0; i < bufLen; i++)
        {
            if ((buf[i] & 0x80) != 0)         /* 1xxxxxxx  */
            {
                if ((buf[i] & 0x40) == 0)     /* 10xxxxxx never 1st byte */
                    return false;
                else                          /* 11xxxxxx begins UTF-8 */
                {
                    int following;

                    if ((buf[i] & 0x20) == 0) {             /* 110xxxxx */
                            following = 1;
                    } else if ((buf[i] & 0x10) == 0) {      /* 1110xxxx */
                            following = 2;
                    } else if ((buf[i] & 0x08) == 0) {      /* 11110xxx */
                            following = 3;
                    } else if ((buf[i] & 0x04) == 0) {      /* 111110xx */
                            following = 4;
                    } else if ((buf[i] & 0x02) == 0) {      /* 1111110x */
                            following = 5;
                    } else
                            return false;

                    for (int n = 0; n < following; n++)
                    {
                        i++;
                        if (i >= bufLen)
                            return false;
                        if ((buf[i] & 0xc0) != 0x80)
                            return false;
                    }
                }
            }
        }
        return true;
    }

    private boolean looks_like_utf16(byte buf[], int bufLen)
    {
        boolean bigend = false;
        if (bufLen < 2)
            return false;

        if (buf[0] == 0xff && buf[1] == 0xfe)
            bigend = true;
        else if (buf[0] == 0xfe && buf[1] == 0xff)
            bigend = true;
        else
            return false;
        return true;
    }

    // does it have internal signature of a tar (Unix Tape Archive) archive?
    // since this includes a computed checksum it has to be tested actively.
    private boolean looks_like_tar(byte buf[], int bufLen)
    {
        // 1. length must be an even multiple of 512 bytes (be sure HOWMANY is!)
        if (bufLen < 1024 || (bufLen % 512) != 0)
            return false;

        // 2. checksum test: bytes 148-155 are octal checksum of first 512-byte block
        // Skip over checksum field and add 8 ASCII Space chars instead.
        int sum = 0;
        for (int i = 0; i < 148; ++i)
            sum += ((int)buf[i]) & 0xff;
        for (int i = 156; i < 512; ++i)
            sum += ((int)buf[i]) & 0xff;
        sum += 8 * 0x20;
        String fileSum = null;
        try
        {
            fileSum = new String(buf, 148, 8, "US-ASCII");
            int nul = fileSum.indexOf('\0');
            if (nul >= 0)
                fileSum = fileSum.substring(0, nul);
            if (sum == Integer.parseInt(fileSum,8))
                return true;
        }
        catch (Exception e)
        {
            if (log.isDebugEnabled())
                log.debug("Got exception parsing TAR Checksum \""+fileSum+"\":",e);
        }

        return false;

       /*** XXX could add this check to distinguish newer format tar?
        *** the PRONOM registry doesn't have a separate format, so no point.
        // check for USTAR
        String ustar = new String(buf, 257, 6, new Charset.forName("US-ASCII"));
        if (ustar.equals("ustar"))
        {
            return ???
        }
       ***/
    }
}
